/*
  File autogenerated by gengetopt version 2.23
  generated with the following command:
  gengetopt --unamed-opts=KEY [OTP] --set-package=oathtool --no-handle-version --no-handle-help --input repro-gengetopt.ggo --file-name oathtool_cmd 

  The developers of gengetopt consider the fixed text that goes in all
  gengetopt output files to be in the public domain:
  we make no copyright claims on it.
*/

/* If we use autoconf.  */
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#ifndef FIX_UNUSED
#define FIX_UNUSED(X) (void) (X) /* avoid warnings for unused params */
#endif

#include <getopt.h>

#include "oathtool_cmd.h"

const char *gengetopt_args_info_purpose = "Generate and validate OATH one-time passwords.  KEY and OTP is the string '-'\nto read from standard input, '@FILE' to read from indicated filename, or a hex\nencoded value (not recommended on multi-user systems).";

const char *gengetopt_args_info_usage = "Usage: oathtool [OPTION]... [KEY [OTP]]...";

const char *gengetopt_args_info_versiontext = "";

const char *gengetopt_args_info_description = "";

const char *gengetopt_args_info_help[] = {
  "  -h, --help                    Print help and exit",
  "  -V, --version                 Print version and exit",
  "      --hotp                    use event-based HOTP mode  (default=on)",
  "      --totp[=MODE]             use time-variant TOTP mode (values \"SHA1\",\n                                  \"SHA256\", or \"SHA512\")  (default=`SHA1')",
  "  -b, --base32                  use base32 encoding of KEY instead of hex\n                                  (default=off)",
  "  -c, --counter=COUNTER         HOTP counter value",
  "  -s, --time-step-size=DURATION TOTP time-step duration  (default=`30s')",
  "  -S, --start-time=TIME         when to start counting time steps for TOTP\n                                  (default=`1970-01-01 00:00:00 UTC')",
  "  -N, --now=TIME                use this time as current time for TOTP\n                                  (default=`now')",
  "  -d, --digits=DIGITS           number of digits in one-time password",
  "  -w, --window=WIDTH            number of additional OTPs to generate or\n                                  validate against",
  "  -v, --verbose                 explain what is being done  (default=off)",
    0
};

typedef enum {ARG_NO
  , ARG_FLAG
  , ARG_STRING
  , ARG_INT
  , ARG_LONGLONG
} cmdline_parser_arg_type;

static
void clear_given (struct gengetopt_args_info *args_info);
static
void clear_args (struct gengetopt_args_info *args_info);

static int
cmdline_parser_internal (int argc, char **argv, struct gengetopt_args_info *args_info,
                        struct cmdline_parser_params *params, const char *additional_error);


static char *
gengetopt_strdup (const char *s);

static
void clear_given (struct gengetopt_args_info *args_info)
{
  args_info->help_given = 0 ;
  args_info->version_given = 0 ;
  args_info->hotp_given = 0 ;
  args_info->totp_given = 0 ;
  args_info->base32_given = 0 ;
  args_info->counter_given = 0 ;
  args_info->time_step_size_given = 0 ;
  args_info->start_time_given = 0 ;
  args_info->now_given = 0 ;
  args_info->digits_given = 0 ;
  args_info->window_given = 0 ;
  args_info->verbose_given = 0 ;
}

static
void clear_args (struct gengetopt_args_info *args_info)
{
  FIX_UNUSED (args_info);
  args_info->hotp_flag = 1;
  args_info->totp_arg = gengetopt_strdup ("SHA1");
  args_info->totp_orig = NULL;
  args_info->base32_flag = 0;
  args_info->counter_orig = NULL;
  args_info->time_step_size_arg = gengetopt_strdup ("30s");
  args_info->time_step_size_orig = NULL;
  args_info->start_time_arg = gengetopt_strdup ("1970-01-01 00:00:00 UTC");
  args_info->start_time_orig = NULL;
  args_info->now_arg = gengetopt_strdup ("now");
  args_info->now_orig = NULL;
  args_info->digits_orig = NULL;
  args_info->window_orig = NULL;
  args_info->verbose_flag = 0;
  
}

static
void init_args_info(struct gengetopt_args_info *args_info)
{


  args_info->help_help = gengetopt_args_info_help[0] ;
  args_info->version_help = gengetopt_args_info_help[1] ;
  args_info->hotp_help = gengetopt_args_info_help[2] ;
  args_info->totp_help = gengetopt_args_info_help[3] ;
  args_info->base32_help = gengetopt_args_info_help[4] ;
  args_info->counter_help = gengetopt_args_info_help[5] ;
  args_info->time_step_size_help = gengetopt_args_info_help[6] ;
  args_info->start_time_help = gengetopt_args_info_help[7] ;
  args_info->now_help = gengetopt_args_info_help[8] ;
  args_info->digits_help = gengetopt_args_info_help[9] ;
  args_info->window_help = gengetopt_args_info_help[10] ;
  args_info->verbose_help = gengetopt_args_info_help[11] ;
  
}

void
cmdline_parser_print_version (void)
{
  printf ("%s %s\n",
     (strlen(CMDLINE_PARSER_PACKAGE_NAME) ? CMDLINE_PARSER_PACKAGE_NAME : CMDLINE_PARSER_PACKAGE),
     CMDLINE_PARSER_VERSION);

  if (strlen(gengetopt_args_info_versiontext) > 0)
    printf("\n%s\n", gengetopt_args_info_versiontext);
}

static void print_help_common(void)
{
	size_t len_purpose = strlen(gengetopt_args_info_purpose);
	size_t len_usage = strlen(gengetopt_args_info_usage);

	if (len_usage > 0) {
		printf("%s\n", gengetopt_args_info_usage);
	}
	if (len_purpose > 0) {
		printf("%s\n", gengetopt_args_info_purpose);
	}

	if (len_usage || len_purpose) {
		printf("\n");
	}

	if (strlen(gengetopt_args_info_description) > 0) {
		printf("%s\n\n", gengetopt_args_info_description);
	}
}

void
cmdline_parser_print_help (void)
{
  int i = 0;
  print_help_common();
  while (gengetopt_args_info_help[i])
    printf("%s\n", gengetopt_args_info_help[i++]);
}

void
cmdline_parser_init (struct gengetopt_args_info *args_info)
{
  clear_given (args_info);
  clear_args (args_info);
  init_args_info (args_info);

  args_info->inputs = 0;
  args_info->inputs_num = 0;
}

void
cmdline_parser_params_init(struct cmdline_parser_params *params)
{
  if (params)
    { 
      params->override = 0;
      params->initialize = 1;
      params->check_required = 1;
      params->check_ambiguity = 0;
      params->print_errors = 1;
    }
}

struct cmdline_parser_params *
cmdline_parser_params_create(void)
{
  struct cmdline_parser_params *params = 
    (struct cmdline_parser_params *)malloc(sizeof(struct cmdline_parser_params));
  cmdline_parser_params_init(params);  
  return params;
}

static void
free_string_field (char **s)
{
  if (*s)
    {
      free (*s);
      *s = 0;
    }
}


static void
cmdline_parser_release (struct gengetopt_args_info *args_info)
{
  unsigned int i;
  free_string_field (&(args_info->totp_arg));
  free_string_field (&(args_info->totp_orig));
  free_string_field (&(args_info->counter_orig));
  free_string_field (&(args_info->time_step_size_arg));
  free_string_field (&(args_info->time_step_size_orig));
  free_string_field (&(args_info->start_time_arg));
  free_string_field (&(args_info->start_time_orig));
  free_string_field (&(args_info->now_arg));
  free_string_field (&(args_info->now_orig));
  free_string_field (&(args_info->digits_orig));
  free_string_field (&(args_info->window_orig));
  
  
  for (i = 0; i < args_info->inputs_num; ++i)
    free (args_info->inputs [i]);

  if (args_info->inputs_num)
    free (args_info->inputs);

  clear_given (args_info);
}


static void
write_into_file(FILE *outfile, const char *opt, const char *arg, const char *values[])
{
  FIX_UNUSED (values);
  if (arg) {
    fprintf(outfile, "%s=\"%s\"\n", opt, arg);
  } else {
    fprintf(outfile, "%s\n", opt);
  }
}


int
cmdline_parser_dump(FILE *outfile, struct gengetopt_args_info *args_info)
{
  int i = 0;

  if (!outfile)
    {
      fprintf (stderr, "%s: cannot dump options to stream\n", CMDLINE_PARSER_PACKAGE);
      return EXIT_FAILURE;
    }

  if (args_info->help_given)
    write_into_file(outfile, "help", 0, 0 );
  if (args_info->version_given)
    write_into_file(outfile, "version", 0, 0 );
  if (args_info->hotp_given)
    write_into_file(outfile, "hotp", 0, 0 );
  if (args_info->totp_given)
    write_into_file(outfile, "totp", args_info->totp_orig, 0);
  if (args_info->base32_given)
    write_into_file(outfile, "base32", 0, 0 );
  if (args_info->counter_given)
    write_into_file(outfile, "counter", args_info->counter_orig, 0);
  if (args_info->time_step_size_given)
    write_into_file(outfile, "time-step-size", args_info->time_step_size_orig, 0);
  if (args_info->start_time_given)
    write_into_file(outfile, "start-time", args_info->start_time_orig, 0);
  if (args_info->now_given)
    write_into_file(outfile, "now", args_info->now_orig, 0);
  if (args_info->digits_given)
    write_into_file(outfile, "digits", args_info->digits_orig, 0);
  if (args_info->window_given)
    write_into_file(outfile, "window", args_info->window_orig, 0);
  if (args_info->verbose_given)
    write_into_file(outfile, "verbose", 0, 0 );
  

  i = EXIT_SUCCESS;
  return i;
}

int
cmdline_parser_file_save(const char *filename, struct gengetopt_args_info *args_info)
{
  FILE *outfile;
  int i = 0;

  outfile = fopen(filename, "w");

  if (!outfile)
    {
      fprintf (stderr, "%s: cannot open file for writing: %s\n", CMDLINE_PARSER_PACKAGE, filename);
      return EXIT_FAILURE;
    }

  i = cmdline_parser_dump(outfile, args_info);
  fclose (outfile);

  return i;
}

void
cmdline_parser_free (struct gengetopt_args_info *args_info)
{
  cmdline_parser_release (args_info);
}

/** @brief replacement of strdup, which is not standard */
char *
gengetopt_strdup (const char *s)
{
  char *result = 0;
  if (!s)
    return result;

  result = (char*)malloc(strlen(s) + 1);
  if (result == (char*)0)
    return (char*)0;
  strcpy(result, s);
  return result;
}

int
cmdline_parser (int argc, char **argv, struct gengetopt_args_info *args_info)
{
  return cmdline_parser2 (argc, argv, args_info, 0, 1, 1);
}

int
cmdline_parser_ext (int argc, char **argv, struct gengetopt_args_info *args_info,
                   struct cmdline_parser_params *params)
{
  int result;
  result = cmdline_parser_internal (argc, argv, args_info, params, 0);

  if (result == EXIT_FAILURE)
    {
      cmdline_parser_free (args_info);
      exit (EXIT_FAILURE);
    }
  
  return result;
}

int
cmdline_parser2 (int argc, char **argv, struct gengetopt_args_info *args_info, int override, int initialize, int check_required)
{
  int result;
  struct cmdline_parser_params params;
  
  params.override = override;
  params.initialize = initialize;
  params.check_required = check_required;
  params.check_ambiguity = 0;
  params.print_errors = 1;

  result = cmdline_parser_internal (argc, argv, args_info, &params, 0);

  if (result == EXIT_FAILURE)
    {
      cmdline_parser_free (args_info);
      exit (EXIT_FAILURE);
    }
  
  return result;
}

int
cmdline_parser_required (struct gengetopt_args_info *args_info, const char *prog_name)
{
  FIX_UNUSED (args_info);
  FIX_UNUSED (prog_name);
  return EXIT_SUCCESS;
}


static char *package_name = 0;

/**
 * @brief updates an option
 * @param field the generic pointer to the field to update
 * @param orig_field the pointer to the orig field
 * @param field_given the pointer to the number of occurrence of this option
 * @param prev_given the pointer to the number of occurrence already seen
 * @param value the argument for this option (if null no arg was specified)
 * @param possible_values the possible values for this option (if specified)
 * @param default_value the default value (in case the option only accepts fixed values)
 * @param arg_type the type of this option
 * @param check_ambiguity @see cmdline_parser_params.check_ambiguity
 * @param override @see cmdline_parser_params.override
 * @param no_free whether to free a possible previous value
 * @param multiple_option whether this is a multiple option
 * @param long_opt the corresponding long option
 * @param short_opt the corresponding short option (or '-' if none)
 * @param additional_error possible further error specification
 */
static
int update_arg(void *field, char **orig_field,
               unsigned int *field_given, unsigned int *prev_given, 
               char *value, const char *possible_values[],
               const char *default_value,
               cmdline_parser_arg_type arg_type,
               int check_ambiguity, int override,
               int no_free, int multiple_option,
               const char *long_opt, char short_opt,
               const char *additional_error)
{
  char *stop_char = 0;
  const char *val = value;
  int found;
  char **string_field;
  FIX_UNUSED (field);

  stop_char = 0;
  found = 0;

  if (!multiple_option && prev_given && (*prev_given || (check_ambiguity && *field_given)))
    {
      if (short_opt != '-')
        fprintf (stderr, "%s: `--%s' (`-%c') option given more than once%s\n", 
               package_name, long_opt, short_opt,
               (additional_error ? additional_error : ""));
      else
        fprintf (stderr, "%s: `--%s' option given more than once%s\n", 
               package_name, long_opt,
               (additional_error ? additional_error : ""));
      return 1; /* failure */
    }

  FIX_UNUSED (default_value);
    
  if (field_given && *field_given && ! override)
    return 0;
  if (prev_given)
    (*prev_given)++;
  if (field_given)
    (*field_given)++;
  if (possible_values)
    val = possible_values[found];

  switch(arg_type) {
  case ARG_FLAG:
    *((int *)field) = !*((int *)field);
    break;
  case ARG_INT:
    if (val) *((int *)field) = strtol (val, &stop_char, 0);
    break;
  case ARG_LONGLONG:
#if defined(HAVE_UNSIGNED_LONG_LONG) || defined(HAVE_UNSIGNED_LONG_LONG_INT)
    if (val) *((long long int*)field) = (long long int) strtoull (val, &stop_char, 0);
#else
    if (val) *((long *)field) = (long)strtol (val, &stop_char, 0);
#endif
    break;
  case ARG_STRING:
    if (val) {
      string_field = (char **)field;
      if (!no_free && *string_field)
        free (*string_field); /* free previous string */
      *string_field = gengetopt_strdup (val);
    }
    break;
  default:
    break;
  };

  /* check numeric conversion */
  switch(arg_type) {
  case ARG_INT:
  case ARG_LONGLONG:
    if (val && !(stop_char && *stop_char == '\0')) {
      fprintf(stderr, "%s: invalid numeric value: %s\n", package_name, val);
      return 1; /* failure */
    }
    break;
  default:
    ;
  };

  /* store the original value */
  switch(arg_type) {
  case ARG_NO:
  case ARG_FLAG:
    break;
  default:
    if (value && orig_field) {
      if (no_free) {
        *orig_field = value;
      } else {
        if (*orig_field)
          free (*orig_field); /* free previous string */
        *orig_field = gengetopt_strdup (value);
      }
    }
  };

  return 0; /* OK */
}


int
cmdline_parser_internal (
  int argc, char **argv, struct gengetopt_args_info *args_info,
                        struct cmdline_parser_params *params, const char *additional_error)
{
  int c;	/* Character of the parsed option.  */

  int error_occurred = 0;
  struct gengetopt_args_info local_args_info;
  
  int override;
  int initialize;
  int check_required;
  int check_ambiguity;
  
  package_name = argv[0];
  
  /* TODO: Why is this here? It is not used anywhere. */
  override = params->override;
  FIX_UNUSED(override);

  initialize = params->initialize;
  check_required = params->check_required;

  /* TODO: Why is this here? It is not used anywhere. */
  check_ambiguity = params->check_ambiguity;
  FIX_UNUSED(check_ambiguity);

  if (initialize)
    cmdline_parser_init (args_info);

  cmdline_parser_init (&local_args_info);

  optarg = 0;
  optind = 0;
  opterr = params->print_errors;
  optopt = '?';

  while (1)
    {
      int option_index = 0;

      static struct option long_options[] = {
        { "help",	0, NULL, 'h' },
        { "version",	0, NULL, 'V' },
        { "hotp",	0, NULL, 0 },
        { "totp",	2, NULL, 0 },
        { "base32",	0, NULL, 'b' },
        { "counter",	1, NULL, 'c' },
        { "time-step-size",	1, NULL, 's' },
        { "start-time",	1, NULL, 'S' },
        { "now",	1, NULL, 'N' },
        { "digits",	1, NULL, 'd' },
        { "window",	1, NULL, 'w' },
        { "verbose",	0, NULL, 'v' },
        { 0,  0, 0, 0 }
      };

      c = getopt_long (argc, argv, "hVbc:s:S:N:d:w:v", long_options, &option_index);

      if (c == -1) break;	/* Exit from `while (1)' loop.  */

      switch (c)
        {
        case 'h':	/* Print help and exit.  */
        
        
          if (update_arg( 0 , 
               0 , &(args_info->help_given),
              &(local_args_info.help_given), optarg, 0, 0, ARG_NO,
              check_ambiguity, override, 0, 0,
              "help", 'h',
              additional_error))
            goto failure;
          cmdline_parser_free (&local_args_info);
          return 0;
        
          break;
        case 'V':	/* Print version and exit.  */
        
        
          if (update_arg( 0 , 
               0 , &(args_info->version_given),
              &(local_args_info.version_given), optarg, 0, 0, ARG_NO,
              check_ambiguity, override, 0, 0,
              "version", 'V',
              additional_error))
            goto failure;
          cmdline_parser_free (&local_args_info);
          return 0;
        
          break;
        case 'b':	/* use base32 encoding of KEY instead of hex.  */
        
        
          if (update_arg((void *)&(args_info->base32_flag), 0, &(args_info->base32_given),
              &(local_args_info.base32_given), optarg, 0, 0, ARG_FLAG,
              check_ambiguity, override, 1, 0, "base32", 'b',
              additional_error))
            goto failure;
        
          break;
        case 'c':	/* HOTP counter value.  */
        
        
          if (update_arg( (void *)&(args_info->counter_arg), 
               &(args_info->counter_orig), &(args_info->counter_given),
              &(local_args_info.counter_given), optarg, 0, 0, ARG_LONGLONG,
              check_ambiguity, override, 0, 0,
              "counter", 'c',
              additional_error))
            goto failure;
        
          break;
        case 's':	/* TOTP time-step duration.  */
        
        
          if (update_arg( (void *)&(args_info->time_step_size_arg), 
               &(args_info->time_step_size_orig), &(args_info->time_step_size_given),
              &(local_args_info.time_step_size_given), optarg, 0, "30s", ARG_STRING,
              check_ambiguity, override, 0, 0,
              "time-step-size", 's',
              additional_error))
            goto failure;
        
          break;
        case 'S':	/* when to start counting time steps for TOTP.  */
        
        
          if (update_arg( (void *)&(args_info->start_time_arg), 
               &(args_info->start_time_orig), &(args_info->start_time_given),
              &(local_args_info.start_time_given), optarg, 0, "1970-01-01 00:00:00 UTC", ARG_STRING,
              check_ambiguity, override, 0, 0,
              "start-time", 'S',
              additional_error))
            goto failure;
        
          break;
        case 'N':	/* use this time as current time for TOTP.  */
        
        
          if (update_arg( (void *)&(args_info->now_arg), 
               &(args_info->now_orig), &(args_info->now_given),
              &(local_args_info.now_given), optarg, 0, "now", ARG_STRING,
              check_ambiguity, override, 0, 0,
              "now", 'N',
              additional_error))
            goto failure;
        
          break;
        case 'd':	/* number of digits in one-time password.  */
        
        
          if (update_arg( (void *)&(args_info->digits_arg), 
               &(args_info->digits_orig), &(args_info->digits_given),
              &(local_args_info.digits_given), optarg, 0, 0, ARG_INT,
              check_ambiguity, override, 0, 0,
              "digits", 'd',
              additional_error))
            goto failure;
        
          break;
        case 'w':	/* number of additional OTPs to generate or validate against.  */
        
        
          if (update_arg( (void *)&(args_info->window_arg), 
               &(args_info->window_orig), &(args_info->window_given),
              &(local_args_info.window_given), optarg, 0, 0, ARG_INT,
              check_ambiguity, override, 0, 0,
              "window", 'w',
              additional_error))
            goto failure;
        
          break;
        case 'v':	/* explain what is being done.  */
        
        
          if (update_arg((void *)&(args_info->verbose_flag), 0, &(args_info->verbose_given),
              &(local_args_info.verbose_given), optarg, 0, 0, ARG_FLAG,
              check_ambiguity, override, 1, 0, "verbose", 'v',
              additional_error))
            goto failure;
        
          break;

        case 0:	/* Long option with no short option */
          /* use event-based HOTP mode.  */
          if (strcmp (long_options[option_index].name, "hotp") == 0)
          {
          
          
            if (update_arg((void *)&(args_info->hotp_flag), 0, &(args_info->hotp_given),
                &(local_args_info.hotp_given), optarg, 0, 0, ARG_FLAG,
                check_ambiguity, override, 1, 0, "hotp", '-',
                additional_error))
              goto failure;
          
          }
          /* use time-variant TOTP mode (values \"SHA1\", \"SHA256\", or \"SHA512\").  */
          else if (strcmp (long_options[option_index].name, "totp") == 0)
          {
          
          
            if (update_arg( (void *)&(args_info->totp_arg), 
                 &(args_info->totp_orig), &(args_info->totp_given),
                &(local_args_info.totp_given), optarg, 0, "SHA1", ARG_STRING,
                check_ambiguity, override, 0, 0,
                "totp", '-',
                additional_error))
              goto failure;
          
          }
          
          break;
        case '?':	/* Invalid option.  */
          /* `getopt_long' already printed an error message.  */
          goto failure;

        default:	/* bug: option not considered.  */
          fprintf (stderr, "%s: option unknown: %c%s\n", CMDLINE_PARSER_PACKAGE, c, (additional_error ? additional_error : ""));
          abort ();
        } /* switch */
    } /* while */



	FIX_UNUSED(check_required);

  cmdline_parser_release (&local_args_info);

  if ( error_occurred )
    return (EXIT_FAILURE);

  if (optind < argc)
    {
      int i = 0 ;
      int found_prog_name = 0;
      /* whether program name, i.e., argv[0], is in the remaining args
         (this may happen with some implementations of getopt,
          but surely not with the one included by gengetopt) */

      i = optind;
      while (i < argc)
        if (argv[i++] == argv[0]) {
          found_prog_name = 1;
          break;
        }
      i = 0;

      args_info->inputs_num = argc - optind - found_prog_name;
      args_info->inputs =
        (char **)(malloc ((args_info->inputs_num)*sizeof(char *))) ;
      while (optind < argc)
        if (argv[optind++] != argv[0])
          args_info->inputs[ i++ ] = gengetopt_strdup (argv[optind-1]) ;
    }

  return 0;

failure:
  
  cmdline_parser_release (&local_args_info);
  return (EXIT_FAILURE);
}
/* vim: set ft=c noet ts=8 sts=8 sw=8 tw=80 nojs spell : */
